module simplecpu
	#(parameter addrSize = 10)
	(input clk,
	 input reset,
	 input in1, in2, in3, in4,
	 inout [7:0] dataBus,
	 output reg [addrSize-1:0] addrBus,
	 output reg we);

// internal CPU registers
reg [addrSize-1:0] pc;
reg [addrSize-1:0] temp;
reg [2:0] mathop;
reg [1:0] addrmode;
reg [7:0] a;

reg aluflagZ;
reg aluflagC;
reg aluflagV;
reg aluflagN; 
wire [3:0] aluflags = { aluflagZ, aluflagC, aluflagV, aluflagN };

// non-register control/internal signals
reg [addrSize-1:0] pcNext, tempNext;
reg [7:0] aNext;
reg [7:0] cmpResult; // incorrectly infers a latch
reg [2:0] mathopNext;
reg [1:0] addrmodeNext;
reg aluflagZNext;
reg aluflagCNext;
reg aluflagVNext;
reg aluflagNNext; 
reg [7:0] dataOut;
	
// states
reg [3:0] state, stateNext;
localparam [3:0] fetchOp = 4'b0000,
				 jump0 = 4'b0001,
				 jump1 = 4'b0010,
				 jump2 = 4'b0011,
				 skipByte0 = 4'b0100,
				 skipByte1 = 4'b0101,
				 setFlags = 4'b0110,
				 doMath = 4'b0111,
				 fetchAddress0 = 4'b1000,
				 fetchAddress1 = 4'b1001;		

// addressing modes
localparam [1:0] immediate = 2'b00,
			     absolute = 2'b01,
			     indirectY = 2'b10,
			     absoluteX = 2'b11;
			     
// math ops
localparam [2:0] opORA = 3'b000,
				 opAND = 3'b001,
				 opEOR = 3'b010,
				 opADC = 3'b011,
				 opSTA = 3'b100,
				 opLDA = 3'b101,
				 opCMP = 3'b110,
				 opSBC = 3'b111;
				 
always @* begin
	// default combinatorial outputs
	addrBus = pc;
	dataOut = 8'bzzzzzzzz;
	we = 1'b1; 
	pcNext = pc;
	tempNext = temp;
	stateNext = state;
	aluflagZNext = aluflagZ;
	aluflagCNext = aluflagC;
	aluflagVNext = aluflagV;
	aluflagNNext = aluflagN;
	addrmodeNext = addrmode;
	mathopNext = mathop;
	aNext = a;
	
	case (state)
		fetchOp: 
			begin
				pcNext = pc + 1'b1;
				
				casez (dataBus)
					8'b???10000: 
						// conditional branches have the form xxy10000, 
						// where xx is the flag and y is the value to match
						if (aluflags[dataBus[7:6]] == dataBus[5])
							stateNext = jump0; // branch taken
						else
							stateNext = skipByte0; // branch not taken
					8'b??????01: 
						// logic and mathematical ops using the accumulator have the form aaabxb01,
						// where aaa is the operation and bbb is the addressing mode
						// aaa: 0=ORA, AND, EOR, ADC, STA, LDA, CMP, 7=SBC
						// bb: 0=immediate, absolute, (ZP),Y, 3=absolute,X 
						begin
							mathopNext = dataBus[7:5];
							addrmodeNext = {dataBus[4], dataBus[2]}; 
							case (addrmodeNext)
								immediate: // immediate
									stateNext = doMath;
								absolute: // absolute
									stateNext = fetchAddress0;
								indirectY: // (ZP),Y
									stateNext = fetchOp; // TODO
								absoluteX: // absolute,X
									stateNext = fetchOp; // TODO
							endcase
						end
					8'h4C: // absolute jump
						stateNext = jump0;
					8'hB8: // set ALU flags
						stateNext = setFlags;
					8'hFF: // halt adds 7 MCs???
						begin
							pcNext = pc;
							stateNext = fetchOp;
						end
					default:
						stateNext = fetchOp; // treat unrecognized opcodes as NOP
				endcase
			end
		skipByte0: 
			begin
				pcNext = pc + 1'b1;
				stateNext = skipByte1;
			end
		skipByte1:
			begin
				pcNext = pc + 1'b1;
				stateNext = fetchOp;
			end
		jump0:
			begin
				pcNext = pc + 1'b1;
				tempNext[7:0] = dataBus;
				stateNext = jump1;
			end
		jump1:
			begin
				pcNext[addrSize-1:8] = dataBus[addrSize-9:0];
				stateNext = jump2;
			end
		jump2:
			begin
				pcNext[7:0] = temp[7:0];
				stateNext = fetchOp;	
			end		
		setFlags:
			begin
				pcNext = pc + 1'b1;
				{ aluflagZNext, aluflagCNext, aluflagVNext, aluflagNNext } = (aluflags | dataBus[3:0]) & dataBus[7:4];
				stateNext = fetchOp;
			end
		fetchAddress0:
			begin
				// note the behavior here is identical to jump0, but next state differs
				pcNext = pc + 1'b1;
				tempNext[7:0] = dataBus;
				stateNext = fetchAddress1;
			end
		fetchAddress1:
			begin
				pcNext = pc + 1'b1;
				tempNext[addrSize-1:8] = dataBus[addrSize-9:0];
				stateNext = doMath;
			end	
		doMath:
			begin
				pcNext = (addrmode == immediate) ? pc + 1'b1 : pc; // advance PC to skip immediate operand
				addrBus = (addrmode == immediate) ? pc : temp;
				case (mathop)
					opORA:
						begin
							aNext = a | dataBus;
							aluflagNNext = aNext[7];
							aluflagZNext = (aNext == 0);
						end
					opAND:
						begin
							aNext = a & dataBus;
							aluflagNNext = aNext[7];
							aluflagZNext = (aNext == 0);
						end
					opEOR:
						begin
							aNext = a ^ dataBus;
							aluflagNNext = aNext[7];
							aluflagZNext = (aNext == 0);
						end
					opADC: 
						begin
							{ aluflagCNext, aNext } = a + dataBus + aluflagC;
							aluflagNNext = aNext[7];
							aluflagZNext = (aNext == 0);
							aluflagVNext = (~a[7] & ~dataBus[7] & aNext[7]) |
										   (a[7] & dataBus[7] & ~aNext[7]);
						end	
					opSTA: 
						begin
							we = 0;
							dataOut = a;
						end
					opLDA:
						begin
							aNext = dataBus;
							aluflagNNext = aNext[7];
							aluflagZNext = (aNext == 0);
						end
					opCMP:
						begin
							{ aluflagCNext, cmpResult } = a - dataBus;
							aluflagNNext = cmpResult[7];
							aluflagZNext = (cmpResult == 0);
						end	
					opSBC:
						begin
							// the expression: 
							//    { aluflagCNext, aNext } = a + ~dataBus + aluflagC; 
							// does NOT work. Why not? aluflagCNext will be incorrect when using that expression. But how is it 
							// different from the fully-expanded version below?
							{ aluflagCNext, aNext } = a + { ~dataBus[7], ~dataBus[6], ~dataBus[5], ~dataBus[4], ~dataBus[3], ~dataBus[2], ~dataBus[1], ~dataBus[0] } + aluflagC;
							aluflagNNext = aNext[7];
							aluflagZNext = (aNext == 0);
							aluflagVNext = (~a[7] & dataBus[7] & aNext[7]) |
										   (a[7] & ~dataBus[7] & ~aNext[7]);
						end	
				endcase
				stateNext = fetchOp;
			end
		default:
			stateNext = fetchOp;	
	endcase
end

always @(posedge clk or negedge reset) begin
	if (!reset) begin
		state <= fetchOp;
		pc <= 0;
	end
	else begin
		state <= stateNext;
		pc <= pcNext;	
		temp <= tempNext;			
		aluflagZ <= aluflagZNext;
		aluflagC <= aluflagCNext;
		aluflagV <= aluflagVNext;
		aluflagN <= aluflagNNext;
		a <= aNext;
		mathop <= mathopNext;
		addrmode <= addrmodeNext;
	end
end

assign dataBus = dataOut;

endmodule

